#include <WiFi.h>
#include <math.h>
#include "PubSubClient.h"
#include <TimeLib.h>
#include <WiFiUdp.h>
#include <WiFiClientSecure.h>
#include "SPIFFS.h"
#include "ArduinoJson.h"
#include <Update.h>

// Define some constants.
#define SERIAL_BAUD     115200
#define HEATER_OUTPUT   4
#define FAN_OUTPUT      5
#define DEBUG_ON
#define VERSION 4

// Declare lots of variables.
int MQTT_PORT;
char* ssid;
char* password;
char* mqtt_server;
char* mqtt_user;
char* mqtt_pass;
char* ntpServerName;
char* fwUrlBase;
char* binFile;
int timeZone = 0;
time_t oldTime = 0;
long contentLength = 0;
bool isValidContentType = false;
int startMinutes, startHours, stopMinutes, stopHours;
const int NTP_PACKET_SIZE = 48; // NTP time is in the first 48 bytes of message
byte packetBuffer[NTP_PACKET_SIZE]; //buffer to hold incoming & outgoing packets

WiFiUDP Udp;
unsigned int localPort = 8888;
WiFiClientSecure espClient;
PubSubClient client(espClient);
WiFiClient insecureClient;

String fwHost;
int fwPort;

// Define some state variables.
bool initalized = false;
bool have_tz = false;
bool scheduling_enabled = false;
bool fan_enabled = false;
bool heater_enabled = false;
bool unsubbed = false;

// Used for the OTA updating function
String getHeaderValue(String header, String headerName) {
  return header.substring(strlen(headerName.c_str()));
}

// Syncs with an external time server to get the time.
time_t getNtpTime() {
  IPAddress ntpServerIP; // NTP server's ip address

  while (Udp.parsePacket() > 0) ; // discard any previously received packets
  Serial.println("Transmit NTP Request");
  // get a random server from the pool
  WiFi.hostByName(ntpServerName, ntpServerIP);
  Serial.print(ntpServerName);
  Serial.print(": ");
  Serial.println(ntpServerIP);
  sendNTPpacket(ntpServerIP);
  uint32_t beginWait = millis();
  while (millis() - beginWait < 1500) {
    int size = Udp.parsePacket();
    if (size >= NTP_PACKET_SIZE) {
      Serial.println("Receive NTP Response");
      Udp.read(packetBuffer, NTP_PACKET_SIZE);  // read packet into the buffer
      unsigned long secsSince1900;
      // convert four bytes starting at location 40 to a long integer
      secsSince1900 =  (unsigned long)packetBuffer[40] << 24;
      secsSince1900 |= (unsigned long)packetBuffer[41] << 16;
      secsSince1900 |= (unsigned long)packetBuffer[42] << 8;
      secsSince1900 |= (unsigned long)packetBuffer[43];
      return secsSince1900 - 2208988800UL + timeZone * SECS_PER_HOUR;
    }
  }
  Serial.println("No NTP Response :-(");
  return 0; // return 0 if unable to get the time
}

// send an NTP request to the time server at the given address
void sendNTPpacket(IPAddress &address) {
  // set all bytes in the buffer to 0
  memset(packetBuffer, 0, NTP_PACKET_SIZE);
  // Initialize values needed to form NTP request
  // (see URL above for details on the packets)
  packetBuffer[0] = 0b11100011;   // LI, Version, Mode
  packetBuffer[1] = 0;     // Stratum, or type of clock
  packetBuffer[2] = 6;     // Polling Interval
  packetBuffer[3] = 0xEC;  // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12] = 49;
  packetBuffer[13] = 0x4E;
  packetBuffer[14] = 49;
  packetBuffer[15] = 52;
  // all NTP fields have been given values, now
  // you can send a packet requesting a timestamp:
  Udp.beginPacket(address, 123); //NTP requests are to port 123
  Udp.write(packetBuffer, NTP_PACKET_SIZE);
  Udp.endPacket();
}

// A function to perform the MQTT connection.
void mqtt_connect() {
  while (!client.connected()) {
    Serial.print("Attempting MQTT connection...");
    if (client.connect("dryerClient",mqtt_user,mqtt_pass)) {
      Serial.println("connected");
      client.subscribe("heater/+");
      client.subscribe("fan/+");
      client.subscribe("schedule/+");
      client.subscribe("dryer/heartbeat");
      client.subscribe("dryer/version");
    } else {
      Serial.print("failed, rc=");
      Serial.print(client.state());
      Serial.println(" try again in 5 seconds");
      delay(5000);
    }
  }
}

// Function that connects to WiFi.
void wifi_connect() {
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);

  Serial.print("Connecting to network ");
  Serial.print(ssid);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println(" Connected!");
}

// Function that enables/disables the fan based on the 'enable' parameter.
// It outputs to the console, toggles the digital pin, and sends an MQTT message
// after the fan has been enabled/disabled.
void fanEnable(bool enable) {
  if(enable) {
    Serial.println("Turning fan on.");
    digitalWrite(FAN_OUTPUT, HIGH);
    client.publish("fan/enable/success", "true");
  } else {
    Serial.println("Turning fan off.");
    digitalWrite(FAN_OUTPUT, LOW);
    client.publish("fan/disable/success", "true");
    if(heater_enabled) {
      heaterEnable(false);
    }
  }
  fan_enabled = enable;
}

// Same as the fanEnable function, just for the heater.
void heaterEnable(bool enable) {
  if(enable) {
    Serial.println("Turning heater on.");
    digitalWrite(HEATER_OUTPUT, HIGH);
    client.publish("heater/enable/success", "true");
    if(fan_enabled == false) {
      fanEnable(true);
    }
  } else {
    Serial.println("Turning heater off.");
    digitalWrite(HEATER_OUTPUT, LOW);
    client.publish("heater/disable/success", "true");
  }
  heater_enabled = enable;
}

// Callback that processes messages as they come in.
void callback(char *topic, byte* message, unsigned int length) {
  Serial.print("Message arrived on topic: ");
  Serial.print(topic);
  Serial.print(". Message: ");
  String messageTemp;

  for (int i = 0; i < length; i++) {
    Serial.print((char)message[i]);
    messageTemp += (char)message[i];
  }
  Serial.println();

  // Topics that enable/disable the heater and fan.
  if (String(topic) == "heater/enable") {
    if(messageTemp == "false") {
      heaterEnable(false);
    } else if(messageTemp == "true") {
      heaterEnable(true);
    }
  } else if(String(topic) == "fan/enable") {
    if(messageTemp == "false") {
      fanEnable(false);
    } else if(messageTemp == "true") {
      fanEnable(true);
    }
    // This is a topic that sends what the current timezone offset is.
    // Once it's received, we can start our scheduler.
  } else if(String(topic) == "schedule/tz") {
    timeZone = messageTemp.toInt() * -1 / 60;
    Serial.print("Setting new timezone offset to: ");
    Serial.println(timeZone);
    setTime(getNtpTime());
    have_tz = true;
    // These two topics send us a new start/stop time.
  } else if(String(topic) == "schedule/start") {
    int epoch = messageTemp.toInt() / 1000;
    startHours = floor(epoch / 3600.0);
    startMinutes = round((((double)epoch / 3600.0) - startHours) * 60.0);
    Serial.print("Setting new schedule start time: ");
    Serial.print(startHours);
    Serial.print(":");
    Serial.println(startMinutes);
  } else if(String(topic) == "schedule/stop") {
    int epoch = messageTemp.toInt() / 1000;
    stopHours = floor(epoch / 3600.0);
    stopMinutes = round((((double)epoch / 3600.0) - stopHours) * 60.0);
    Serial.print("Setting new schedule stop time: ");
    Serial.print(stopHours);
    Serial.print(":");
    Serial.println(stopMinutes);
    // A topic used for enabling/disabling the scheduler.
  } else if(String(topic) == "schedule/enable") {
    if(messageTemp == "false") {
      scheduling_enabled = false;
    } else if(messageTemp == "true") {
      scheduling_enabled = true;
    }
  } else if(String(topic) == "dryer/heartbeat") {
    Serial.print("Responding to heartbeat with hash ");
    Serial.println(messageTemp);
    client.publish("dryer/heartbeat/response", messageTemp.c_str());
  } else if(String(topic) == "dryer/version") {
    int server_version = messageTemp.toInt();
    if(server_version > VERSION) {
      Serial.println("Firmware out of date. Updating.");
      fwUpdate();
    } else {
      Serial.println("We have the latest firmware!");
      client.publish("dryer/update/current_version", String(VERSION).c_str());
    }
  }
}

void fwUpdate() {
  Serial.println("Connecting to " + fwHost);
  if (insecureClient.connect(fwHost.c_str(), fwPort)) {
    Serial.println("Connected!");
    Serial.println("Fetching Bin: " + String(binFile));
    insecureClient.print(String("GET ") + binFile + " HTTP/1.1\r\n" +
    "Host: " + fwHost + "\r\n" +
    "Cache-Control: no-cache\r\n" +
    "Connection: close\r\n\r\n");
    
    unsigned long timeout = millis();
    while (insecureClient.available() == 0) {
      if (millis() - timeout > 5000) {
        Serial.println("Client Timeout !");
        insecureClient.stop();
        return;
      }
    }
    
    while(insecureClient.available()) {
      String line = insecureClient.readStringUntil('\n');
      line.trim();
      if (!line.length()) {
        //headers ended
        break; // and get the OTA started
      }

      if (line.startsWith("HTTP/1.1")) {
        if (line.indexOf("200") < 0) {
          Serial.println("Got a non 200 status code from server. Exiting OTA Update.");
          break;
        }
      }
      
      if (line.startsWith("Content-Length: ")) {
        contentLength = atol((getHeaderValue(line, "Content-Length: ")).c_str());
        Serial.println("Got " + String(contentLength) + " bytes from server");
      }

      // Next, the content type
      if (line.startsWith("Content-Type: ")) {
        String contentType = getHeaderValue(line, "Content-Type: ");
        Serial.println("Got " + contentType + " payload.");
        if (contentType == "application/octet-stream") {
          isValidContentType = true;
        }
      }
    }
  } else {
    Serial.println("Connection to " + String(fwHost) + " failed. Please check your setup");
  }

  Serial.println("contentLength : " + String(contentLength) + ", isValidContentType : " + String(isValidContentType));

  if (contentLength && isValidContentType) {
    // Check if there is enough to OTA Update
    bool canBegin = Update.begin(contentLength);

    // If yes, begin
    if (canBegin) {
      Serial.println("Begin OTA. This may take 2 - 5 mins to complete. Things might be quite for a while.. Patience!");
      // No activity would appear on the Serial monitor
      // So be patient. This may take 2 - 5mins to complete
      size_t written = Update.writeStream(insecureClient);

      if (written == contentLength) {
        Serial.println("Written : " + String(written) + " successfully");
      } else {
        Serial.println("Written only : " + String(written) + "/" + String(contentLength) + ". Retry?" );
        // retry??
        // execOTA();
      }

      if (Update.end()) {
        Serial.println("OTA done!");
        if (Update.isFinished()) {
          Serial.println("Update successfully completed. Rebooting.");
          ESP.restart();
        } else {
          Serial.println("Update not finished? Something went wrong!");
        }
      } else {
        Serial.println("Error Occurred. Error #: " + String(Update.getError()));
      }
    } else {
      // not enough space to begin OTA
      // Understand the partitions and
      // space availability
      Serial.println("Not enough space to begin OTA");
      insecureClient.flush();
    }
  } else {
    Serial.println("There was no content in the response");
    insecureClient.flush();
  }
}

void setup() {
  // Initialize the heater and fan outputs and start with them turned off.
  pinMode(HEATER_OUTPUT, OUTPUT);
  pinMode(FAN_OUTPUT, OUTPUT);

  heaterEnable(false);
  fanEnable(false);

  // Connect to serial, SPIFFS, WiFi, MQTT and NTP.
  Serial.begin(SERIAL_BAUD);
  
  if(!SPIFFS.begin(true)) {
    Serial.println("Error mounting SPIFFS.");
  }
  
  File jsonFile = SPIFFS.open("/config.json");
  
  if(!jsonFile) {
    Serial.println("Error opening JSON config file.");
  }
  String jsonString = jsonFile.readString();
  jsonFile.close();
  
  StaticJsonDocument<1024> configData;
  
  DeserializationError err = deserializeJson(configData, jsonString);
  
  if(err) {
    Serial.println("Error parsing JSON.");
    Serial.println(err.c_str());
  }
  
  MQTT_PORT     = configData["mqttPort"];
  
  // All the strings the JSON parser gives are const char*'s, so they have to
  // be converted.
  const char* ssid_tmp          = configData["wifiSsid"];
  const char* password_tmp      = configData["wifiPassword"];
  const char* mqtt_server_tmp   = configData["mqttServer"];
  const char* mqtt_user_tmp     = configData["mqttUser"];
  const char* mqtt_pass_tmp     = configData["mqttPass"];
  const char* ntpServerName_tmp = configData["ntpServer"];
  const char* binFile_tmp     = configData["fw_bin_file"];
  const char* fwHost_tmp = configData["fw_server"];
  fwPort = configData["fw_port"];
  
  ssid = strdup(ssid_tmp);
  password = strdup(password_tmp);
  mqtt_server = strdup(mqtt_server_tmp);
  mqtt_user = strdup(mqtt_user_tmp);
  mqtt_pass = strdup(mqtt_pass_tmp);
  ntpServerName = strdup(ntpServerName_tmp);
  binFile = strdup(binFile_tmp);
  fwHost = String(strdup(fwHost_tmp));
  
  wifi_connect();
  
  File file = SPIFFS.open("/root.cer");
  
  if(!file) {
    Serial.println("Error opening the file.");
  }
  
  String ca_cert = file.readString();
  
  Serial.println("CA Root certificate: ");
  Serial.print(ca_cert);

  char *dest;
  
  dest = (char *)malloc(sizeof(char) * (ca_cert.length()+1));
  strcpy(dest, ca_cert.c_str());
  
  espClient.setCACert(dest);
  file.close();
  
  Serial.print("IP Address: ");
  Serial.println(WiFi.localIP());
  Serial.println("Starting UDP");
  Udp.begin(localPort);
  Serial.println("waiting for sync");
  setSyncProvider(getNtpTime);
  setSyncInterval(300);

  client.setServer(mqtt_server, MQTT_PORT);
  client.setCallback(callback);
}

// Check if a time is within our scheduling interval
bool inTimeRange(time_t current_time) {
  int currentHours = hour(current_time);
  int currentMinutes = minute(current_time);
  
  if(currentHours > stopHours || (currentMinutes >= stopMinutes && currentHours >= stopHours)) {
    return false;
  } else if(currentHours > startHours || (currentMinutes >= startMinutes && currentHours >= startHours)) {
    return true;
  } else {
    return false;
  }
}

void loop() {
  // Once we connect to MQTT, turn things off.
  if (!client.connected()) {
    mqtt_connect();
    heaterEnable(false);
    fanEnable(false);
  }

  client.loop();

  // Tell the MQTT server that we connected so that it sends us Timezone 
  // information and initial values for start/stop times.
  if(!initalized) {
    client.publish("dryer/connected", "connected");
    initalized = true;
  }

  // Once we have timezone information and scheduling is enabled, operate on the scheduler.
  if(have_tz && scheduling_enabled) {
    // Unsubscribe from fan and heater topics so that we ignore the switches and operate only on the scheduler.
    if(!unsubbed) {
      client.unsubscribe("fan/+");
      client.unsubscribe("heater/+");
      unsubbed = true;
    }
    
    // See if we're within the scheduling period and enable/disable the heater and fan accordingly.
    if(inTimeRange(now())) {
      if(!heater_enabled) {
        Serial.println("Starting heater according to schedule.");
        heaterEnable(true);
      }

      if(!fan_enabled) {
        Serial.println("Starting heater according to schedule.");
        fanEnable(true);
      }
    } else {
      if(heater_enabled) {
        Serial.println("Stopping heater according to schedule.");
        heaterEnable(false);
      }

      if(fan_enabled) {
        Serial.println("Stopping fan according to schedule.");
        fanEnable(false);
      }
    }
  } else if(!scheduling_enabled && unsubbed) {
    // Once scheduling is disabled, resubscribe to the manual control switches.
    heaterEnable(false);
    fanEnable(false);
    unsubbed = false;
    client.subscribe("fan/+");
    client.subscribe("heater/+");
  }
}
